; ==================================================================================================
; Title:   Autobuffer.inc
; Author:  G. Friedrich, original idea from MichaelW.
; Version: 1.0.0
; Purpose: ObjAsm32 support of Autobuffer objects.
; Link:    http://www.masm32.com/board/index.php?topic=16378.0
; Notes:   Version 1.0.0, April 2011
;            - First relese.
; ==================================================================================================


; 
; Object:  Autobuffer
; Purpose: This object is designed to serve as an autoexpanding buffer up to a limit specified
;          when the object is created. It differs from the system stack in the growing direction
;          which is positiv on this object.
;          The Startup static method installs the UnhandledExceptionFilter for all Autobuffer
;          instances. This method is called automatically when SysInit is invoked.

Object Autobuffer, AutobufferID, Primer
  RedefineMethod    Done
  RedefineMethod    Init,         Pointer, dword, dword
  StaticMethod      Startup

  DefineVariable    pBuffer,                Pointer,      NULL
  DefineVariable    pGuardPage,             Pointer,      NULL
  DefineVariable    pLastPage,              Pointer,      NULL
  DefineVariable    dPageSize,              dword,        0
  DefineVariable    pPrevExceptionFilter,   Pointer,      NULL
  DefineVariable    ObjLock,                ObjectLock,   {0, 0}
ObjectEnd


; ==================================================================================================

if IMPLEMENT

.data

Autobuffer_pPrevExceptionFilter  Pointer NULL

.code

EXCEPTION_GUARD_PAGE equ 080000001h

; 
; Procedure: GuardPageExceptionFilter
; Purpose:   Unhandled exception filter for Autobuffer objects.
; Arguments: Arg1: -> EXCEPTION_POINTERS
; Return:    eax = EXCEPTION_CONTINUE_EXECUTION or EXCEPTION_CONTINUE_SEARCH

GuardPageExceptionFilter proc pEXCEPTION_POINTERS:ptr EXCEPTION_POINTERS
    mov edx, pEXCEPTION_POINTERS
    mov ecx, [edx].EXCEPTION_POINTERS.pExceptionRecord
    ;Handle the guard page exception and let the system handle any other exception.
;    DbgHex [ecx].EXCEPTION_RECORD.ExceptionCode
    .if [ecx].EXCEPTION_RECORD.ExceptionCode == EXCEPTION_GUARD_PAGE
      ;At this point the system has cleared the PAGE_GUARD modifier so the page that was formerly
      ;the guard page is now a normal PAGE_READWRITE page, and we need to commit a new guard page.
;      DbgHex [ecx].EXCEPTION_RECORD.NumberParameters
;      DbgHex [ecx + 0].EXCEPTION_RECORD.ExceptionInformation    ;0 = read, 1 = write violation
;      DbgHex [ecx + 4].EXCEPTION_RECORD.ExceptionInformation    ;address of access violation
      mov eax, [ecx + 4].EXCEPTION_RECORD.ExceptionInformation
      and eax, 0FFFFF000h
      push esi
      SetObject esi, Autobuffer, dword ptr [eax]
      m2z dword ptr [eax]                               ;Set to zero as expected
      .if esi != NULL
        push eax

        ;Check if the object is an Autobuffer
        invoke GetObjectID, esi
        .if eax == AutobufferID

          ;Lock access to avoid data corruption
          LockObjectAccess [esi]
          pop eax

	        ;Check if the exception comes from our buffer
	        .if eax == [esi].pGuardPage
	          .if eax != [esi].pLastPage
		          add eax, [esi].dPageSize
		          mov [esi].pGuardPage, eax
		          invoke VirtualAlloc, eax, [esi].dPageSize, MEM_COMMIT, PAGE_READWRITE
		          .if eax != NULL
		            mov [eax], esi
		            invoke VirtualAlloc, [esi].pGuardPage, [esi].dPageSize, MEM_COMMIT, \
		                                 PAGE_READWRITE or PAGE_GUARD
		          .else
		            OCall esi.ErrorReport, NULL, ABF_OUT_OF_MEMORY
		          .endif
		          UnlockObjectAccess [esi]
		          ;This return value will cause execution to continue from the point of the exception.
		          mov eax, EXCEPTION_CONTINUE_EXECUTION
		          pop esi
		          ret
	          .else
	            OCall esi.ErrorReport, NULL, ABF_MEMORY_EXHAUSTED
		        .endif
	        .endif
          UnlockObjectAccess [esi]
	      .else
	        add esp, 4
	      .endif
	    .endif
      pop esi
    .endif

    ;Call the previous exception filter (if any)
    mov eax, [Autobuffer_pPrevExceptionFilter]
    .if eax != NULL
	    push pEXCEPTION_POINTERS
	    call eax
	  .endif                                               ;eax = 0 => EXCEPTION_CONTINUE_SEARCH
    ReleaseObject
    ret
GuardPageExceptionFilter endp

; 
; Method:    Autobuffer.Init
; Purpose:   Initializes the Autobuffer object.
; Arguments: Arg1: -> Owner object.
;            Arg2: Initial number of pages. Zero is allowed.
;            Arg3: Maximal number of pages. Must be bigger than the initial number of pages.
; Return:    eax -> Buffer or NULL if failed.

Method Autobuffer.Init, uses esi, pOwner:Pointer, dPageCnt:dword, dMaxPages:dword
    local sysinfo:SYSTEM_INFO

    SetObject esi
    ACall esi.Init, pOwner

    ;Check parameter correctness
    mov eax, dPageCnt
    .if eax < dMaxPages
      ;Get and store the page size (normally 4096 bytes).
      invoke GetSystemInfo, addr sysinfo
      m2m [esi].dPageSize, sysinfo.dwPageSize, eax

      ;Calc buffer limit in bytes.
      inc dMaxPages                                     ;Include last guard page
      mul dMaxPages                                     ;(MaxPages + 1) * PageSize
      push eax

      ;Reserve the memory and store the base address. If any allocation fails here, return NULL.
      invoke VirtualAlloc, 0, eax, MEM_RESERVE, PAGE_NOACCESS
      mov [esi].pBuffer, eax
      pop ecx

      .if eax != NULL
        ;Calc last possible guard page
        add ecx, eax
        mov eax, [esi].dPageSize
        sub ecx, eax
        mov [esi].pLastPage, ecx

        mul dPageCnt                                    ;PageCnt * PageSize

        mov ecx, [esi].pBuffer
        add ecx, eax
        mov [esi].pGuardPage, ecx
        add eax, [esi].dPageSize

        ;Commit page count + 1.
        invoke VirtualAlloc, [esi].pBuffer, eax , MEM_COMMIT, PAGE_READWRITE
        .if eax != NULL
          ;Save object instance on it.
          mov edx, [esi].pGuardPage
          mov [edx], esi
          ;Mark it as guard page.
          invoke VirtualAlloc, edx, [esi].dPageSize , MEM_COMMIT, PAGE_READWRITE or PAGE_GUARD

          ;Return a pointer to the buffer.
          mov eax, [esi].pBuffer
        .endif
      .endif
    .else
      xor eax, eax
    .endif
    DbgObject Self::Autobuffer
MethodEnd

; 
; Method:    Autobuffer.Done
; Purpose:   Finalizes the Autobuffer object.
; Arguments: None.
; Return:    Nothing.

Method Autobuffer.Done, uses esi
    SetObject esi
    invoke VirtualFree, [esi].pBuffer, 0, MEM_RELEASE
    ACall esi.Done
MethodEnd

; 
; Method:    Autobuffer.Startup
; Purpose:   Replace the existing top-level exception filter with our guard page exception filter,
;            so when an unhandled exception reaches the system UnhandledExceptionFilter function,
;            and the process is not being debugged, the function will call our filter procedure.
;            Since SysInit is called at application start, there aren't any multithreading issues to
;            consider.
; Arguments: None.
; Return:    Nothing.

Method Autobuffer.Startup
    invoke SetUnhandledExceptionFilter, offset GuardPageExceptionFilter
    mov Autobuffer_pPrevExceptionFilter, eax            ;Store this address in a static variable
MethodEnd

endif